@@ -84,18 +84,29 @@ static inline void list_sort_selectionsort(list_t *l, int versus,
 
 static void *list_get_minmax(const list_t *l, int versus);
 
-static inline struct list_entry_s *list_findpos(const list_t *l, unsigned int posstart);
+static inline struct list_entry_s *list_findpos(const list_t *l, int posstart);
 
 /* list initialization */
 int list_init(list_t *l) {
     if (l == NULL) return -1;
 
-    l->head = l->mid = l->tail = NULL;
     l->numels = 0;
+
+    /* head/tail sentinels and mid pointer */
+    l->head_sentinel = (struct list_entry_s *)malloc(sizeof(struct list_entry_s));
+    l->tail_sentinel = (struct list_entry_s *)malloc(sizeof(struct list_entry_s));
+    l->head_sentinel->next = l->tail_sentinel;
+    l->tail_sentinel->prev = l->head_sentinel;
+    l->head_sentinel->prev = l->tail_sentinel->next = l->mid = NULL;
+    l->head_sentinel->data = l->tail_sentinel->data = NULL;
+
+    /* iteration attributes */
     l->iter_active = 0;
     l->iter_pos = 0;
     l->iter_curentry = NULL;
-    l->spareels = malloc(SIMCLIST_MAX_SPARE_ELEMS * sizeof(struct list_entry_s *));
+
+    /* free-list attributes */
+    l->spareels = (struct list_entry_s **)malloc(SIMCLIST_MAX_SPARE_ELEMS * sizeof(struct list_entry_s *));
     l->spareelsnum = 0;
 
 #ifdef SIMCLIST_WITH_THREADS
@@ -118,7 +129,8 @@ void list_destroy(list_t *l) {
         free(l->spareels[i]);
     }
     free(l->spareels);
-    assert(list_repOk(l));
+    free(l->head_sentinel);
+    free(l->tail_sentinel);
 }
 
 int list_attributes_setdefaults(list_t *l) {
@@ -187,7 +199,8 @@ void *list_get_min(const list_t *l) {
     return list_get_minmax(l, -1);
 }
 
-/* return the min (versus < 0) or max value (v > 0) in l */
+/* REQUIRES {list->numels >= 1}
+ * return the min (versus < 0) or max value (v > 0) in l */
 static void *list_get_minmax(const list_t *l, int versus) {
     void *curminmax;
     struct list_entry_s *s;
@@ -195,8 +208,8 @@ static void *list_get_minmax(const list_t *l, int versus) {
     if (l->attrs.comparator == NULL || l->numels == 0)
         return NULL;
     
-    curminmax = l->head->data;
-    for (s = l->head->next; s != NULL; s = s->next) {
+    curminmax = l->head_sentinel->next->data;
+    for (s = l->head_sentinel->next->next; s != l->tail_sentinel; s = s->next) {
         if (l->attrs.comparator(curminmax, s->data) * versus > 0)
             curminmax = s->data;
     }
@@ -205,17 +218,18 @@ static void *list_get_minmax(const list_t *l, int versus) {
 }
 
 /* set tmp to point to element at index posstart in l */
-static inline struct list_entry_s *list_findpos(const list_t *l, unsigned int posstart) {
+static inline struct list_entry_s *list_findpos(const list_t *l, int posstart) {
     struct list_entry_s *ptr;
     float x;
-    unsigned int i;
+    int i;
 
-    if (posstart >= l->numels) return NULL;
+    /* accept 1 slot overflow for fetching head and tail sentinels */
+    if (posstart < -1 || posstart > (int)l->numels) return NULL;
 
     x = (float)(posstart+1) / l->numels;
     if (x <= 0.25) {
         /* first quarter: get to posstart from head */
-        for (i = 0, ptr = l->head; i < posstart; ptr = ptr->next, i++);
+        for (i = -1, ptr = l->head_sentinel; i < posstart; ptr = ptr->next, i++);
     } else if (x < 0.5) {
         /* second quarter: get to posstart from mid */
         for (i = (l->numels-1)/2, ptr = l->mid; i > posstart; ptr = ptr->prev, i--);
@@ -224,7 +238,7 @@ static inline struct list_entry_s *list_findpos(const list_t *l, unsigned int po
         for (i = (l->numels-1)/2, ptr = l->mid; i < posstart; ptr = ptr->next, i++);
     } else {
         /* fourth quarter: get to posstart from tail */
-        for (i = l->numels-1, ptr = l->tail; i > posstart; ptr = ptr->prev, i--);
+        for (i = l->numels, ptr = l->tail_sentinel; i > posstart; ptr = ptr->prev, i--);
     }
 
     return ptr;
@@ -234,8 +248,7 @@ void *list_extract_at(list_t *l, unsigned int pos) {
     struct list_entry_s *tmp;
     void *data;
     
-    if (l->iter_active) return NULL;
-    if (pos >= l->numels) return NULL;
+    if (l->iter_active || pos >= l->numels) return NULL;
 
     tmp = list_findpos(l, pos);
     data = tmp->data;
@@ -252,15 +265,14 @@ void *list_extract_at(list_t *l, unsigned int pos) {
 int list_insert_at(list_t *l, const void *data, unsigned int pos) {
     struct list_entry_s *lent, *succ, *prec;
     
-    if (l->iter_active) return -1;
-    if (pos > l->numels) return -2;
+    if (l->iter_active || pos > l->numels) return -1;
     
     /* this code optimizes malloc() with a free-list */
     if (l->spareelsnum > 0) {
         lent = l->spareels[l->spareelsnum-1];
         l->spareelsnum--;
     } else {
-        lent = malloc(sizeof(struct list_entry_s));
+        lent = (struct list_entry_s *)malloc(sizeof(struct list_entry_s));
         if (lent == NULL)
             return -1;
     }
@@ -268,48 +280,28 @@ int list_insert_at(list_t *l, const void *data, unsigned int pos) {
     if (l->attrs.copy_data) {
         /* make room for user' data (has to be copied) */
         size_t datalen = l->attrs.meter(data);
-        lent->data = malloc(datalen);
+        lent->data = (struct list_entry_s *)malloc(datalen);
         memcpy(lent->data, data, datalen);
     } else {
         lent->data = (void*)data;
     }
 
-
-    if (l->numels == 0) {       /* list was empty */
-        lent->next = lent->prev = NULL;
-        l->head = l->mid = l->tail = lent;
-        l->numels++;
-        return 1;
-    }
-
-    if (pos == 0) {             /* list not empty, but this is the head */
-        lent->next = l->head;
-        lent->prev = NULL;
-        l->head->prev = lent;
-        l->head = lent;
-        if (! (l->numels % 2)) l->mid = l->mid->prev;   /* now even */
-        l->numels++;
-        return 1;
-    }
-
-    /* elem to be appended in the middle (or tail) of the list */
+    /* actually append element */
     prec = list_findpos(l, pos-1);
     succ = prec->next;
     
     prec->next = lent;
     lent->prev = prec;
     lent->next = succ;
-    if (succ == NULL) {
-        l->tail = lent;
-    } else {
-        succ->prev = lent;
-    }
+    succ->prev = lent;
 
     l->numels++;
 
     /* fix mid pointer */
-    if (l->numels % 2) {    /* now odd */
-        if (pos > (l->numels-1)/2) l->mid = l->mid->next;
+    if (l->numels == 1) { /* first element, set pointer */
+        l->mid = lent;
+    } else if (l->numels % 2) {    /* now odd */
+        if (pos >= (l->numels-1)/2) l->mid = l->mid->next;
     } else {                /* now even */
         if (pos <= (l->numels-1)/2) l->mid = l->mid->prev;
     }
@@ -323,8 +315,7 @@ int list_delete_at(list_t *l, unsigned int pos) {
     struct list_entry_s *delendo;
 
 
-    if (l->iter_active) return -2;
-    if (pos >= l->numels) return -1;
+    if (l->iter_active || pos >= l->numels) return -1;
 
     delendo = list_findpos(l, pos);
 
@@ -344,12 +335,10 @@ int list_delete_range(list_t *l, unsigned int posstart, unsigned int posend) {
     int movedx;
     unsigned int numdel, midposafter;
 
-    if (l->iter_active) return -2;
-    if (posend < posstart) return -1;
-    if (posend >= l->numels) return -1;
+    if (l->iter_active || posend < posstart || posend >= l->numels) return -1;
 
-    tmp = list_findpos(l, posstart);
-    lastvalid = tmp->prev;
+    tmp = list_findpos(l, posstart);    /* first el to be deleted */
+    lastvalid = tmp->prev;              /* last valid element */
 
     numdel = posend - posstart + 1;
     midposafter = (l->numels-1-numdel)/2;
@@ -364,7 +353,7 @@ int list_delete_range(list_t *l, unsigned int posstart, unsigned int posend) {
         for (i = 0; i < (unsigned int)movedx; l->mid = l->mid->prev, i++);
     }
 
-    assert(posstart == 0 || lastvalid != NULL);
+    assert(posstart == 0 || lastvalid != l->head_sentinel);
     i = posstart;
     if (l->attrs.copy_data) {
         /* also free element data */
@@ -390,18 +379,10 @@ int list_delete_range(list_t *l, unsigned int posstart, unsigned int posend) {
             }
         }
     }
-    assert(i == posend+1 && (posend != l->numels || tmp == NULL));
+    assert(i == posend+1 && (posend != l->numels || tmp == l->tail_sentinel));
 
-    if (posstart == 0) {    /* head gone */
-        l->head = tmp;
-    } else {    /* first != NULL */
-        lastvalid->next = tmp;
-    }
-    if (posend == l->numels-1) { /* tail gone */
-        l->tail = lastvalid;
-    } else {    /* tmp != NULL */
-        tmp->prev = lastvalid;
-    }
+    lastvalid->next = tmp;
+    tmp->prev = lastvalid;
 
     l->numels -= posend - posstart + 1;
 
@@ -412,40 +393,44 @@ int list_delete_range(list_t *l, unsigned int posstart, unsigned int posend) {
 
 int list_clear(list_t *l) {
     struct list_entry_s *s;
-    unsigned int i;
 
     if (l->iter_active) return -1;
     
-    if (l->attrs.copy_data) {
-        /* also free user data */
-        for (i = 0, s = l->head; i < l->numels; i++) {
-            s = s->next;
-            if (l->head->data != NULL) free(l->head->data);
-            if (l->spareelsnum < SIMCLIST_MAX_SPARE_ELEMS) {
-                l->spareels[l->spareelsnum++] = l->head;
-            } else {
-                free(l->head);
-            }
-            l->head = s;
+    if (l->attrs.copy_data) {        /* also free user data */
+        /* spare a loop conditional with two loops: spareing elems and freeing elems */
+        for (s = l->head_sentinel->next; l->spareelsnum < SIMCLIST_MAX_SPARE_ELEMS && s != l->tail_sentinel; s = s->next) {
+            /* move elements as spares as long as there is room */
+            if (s->data != NULL) free(s->data);
+            l->spareels[l->spareelsnum++] = s;
         }
-    } else {
-        /* only free element containers */
-        for (i = 0, s = l->head; i < l->numels; i++) {
+        while (s != l->tail_sentinel) {
+            /* free the remaining elems */
+            if (s->data != NULL) free(s->data);
+            free(s);
             s = s->next;
-            if (l->spareelsnum < SIMCLIST_MAX_SPARE_ELEMS) {
-                l->spareels[l->spareelsnum++] = l->head;
-            } else {
-                free(l->head);
-            }
-            l->head = s;
         }
+        l->head_sentinel->next = l->tail_sentinel;
+        l->tail_sentinel->prev = l->head_sentinel;
+    } else { /* only free element containers */
+        /* spare a loop conditional with two loops: spareing elems and freeing elems */
+        for (s = l->head_sentinel->next; l->spareelsnum < SIMCLIST_MAX_SPARE_ELEMS && s != l->tail_sentinel; s = s->next) {
+            /* move elements as spares as long as there is room */
+            l->spareels[l->spareelsnum++] = s;
+        }
+        while (s != l->tail_sentinel) {
+            /* free the remaining elems */
+            free(s);
+            s = s->next;
+        }
+        l->head_sentinel->next = l->tail_sentinel;
+        l->tail_sentinel->prev = l->head_sentinel;
     }
     l->numels = 0;
-    l->head = l->tail = l->mid = NULL;
+    l->mid = NULL;
 
     assert(list_repOk(l));
 
-    return i;
+    return 0;
 }
 
 unsigned int list_size(const list_t *l) {
@@ -462,16 +447,16 @@ int list_find(const list_t *l, const void *data) {
     
     if (l->attrs.comparator != NULL) {
         /* use comparator */
-        for (el = l->head; el != NULL; el = el->next, pos++) {
+        for (el = l->head_sentinel->next; el != l->tail_sentinel; el = el->next, pos++) {
             if (l->attrs.comparator(data, el->data) == 0) break;
         }
     } else {
         /* compare references */
-        for (el = l->head; el != NULL; el = el->next, pos++) {
+        for (el = l->head_sentinel->next; el != l->tail_sentinel; el = el->next, pos++) {
             if (el->data == data) break;
         }
     }
-    if (el == NULL) return -1;
+    if (el == l->tail_sentinel) return -1;
 
     return pos;
 }
@@ -495,39 +480,40 @@ int list_concat(const list_t *l1, const list_t *l2, list_t *dest) {
     if (dest->numels == 0)
         return 0;
 
-    el = dest->head = malloc(sizeof(struct list_entry_s));
-    el->prev = NULL;
-    for (cnt = 0, srcel = l1->head; cnt < l1->numels; cnt++) {
-        el->data = srcel->data;
-        el->next = malloc(sizeof(struct list_entry_s));
+    /* copy list1 */
+    srcel = l1->head_sentinel->next;
+    el = dest->head_sentinel;
+    while (srcel != l1->tail_sentinel) {
+        el->next = (struct list_entry_s *)malloc(sizeof(struct list_entry_s));
         el->next->prev = el;
         el = el->next;
+        el->data = srcel->data;
         srcel = srcel->next;
     }
     dest->mid = el;     /* approximate position (adjust later) */
-    for (cnt = 0, srcel = l2->head; cnt < l2->numels; cnt++) {
-        el->data = srcel->data;
-        el->next = malloc(sizeof(struct list_entry_s));
+    /* copy list 2 */
+    srcel = l2->head_sentinel->next;
+    while (srcel != l2->tail_sentinel) {
+        el->next = (struct list_entry_s *)malloc(sizeof(struct list_entry_s));
         el->next->prev = el;
         el = el->next;
+        el->data = srcel->data;
         srcel = srcel->next;
     }
-    dest->tail = el->prev;
-    el->prev->next = NULL;
-    dest->spareels[0] = el;
-    dest->spareelsnum = 1;
+    el->next = dest->tail_sentinel;
+    dest->tail_sentinel->prev = el;
     
     /* fix mid pointer */
     err = l2->numels - l1->numels;
-    if ((err-1)/2 > 0) {        /* correct pos RIGHT (err-1)/2 moves */
-        err = (err-1)/2;
+    if ((err+1)/2 > 0) {        /* correct pos RIGHT (err-1)/2 moves */
+        err = (err+1)/2;
         for (cnt = 0; cnt < (unsigned int)err; cnt++) dest->mid = dest->mid->next;
-    } else if ((err/2)-1 < 0) { /* correct pos LEFT (err/2)-1 moves */
-        err = -(err/2-1);
+    } else if (err/2 < 0) { /* correct pos LEFT (err/2)-1 moves */
+        err = -err/2;
         for (cnt = 0; cnt < (unsigned int)err; cnt++) dest->mid = dest->mid->prev;
     }
  
-    assert(list_repOk(dest));
+    assert(!(list_repOk(l1) && list_repOk(l2)) || list_repOk(dest));
 
     return 0;
 }
@@ -539,7 +525,7 @@ int list_sort(list_t *l, int versus) {
     if (l->numels <= 1)
         return 0;
     srandom(time(NULL)*time(NULL));
-    list_sort_quicksort(l, versus, 0, l->head, l->numels-1, l->tail);
+    list_sort_quicksort(l, versus, 0, l->head_sentinel->next, l->numels-1, l->tail_sentinel->prev);
     assert(list_repOk(l));
     return 0;
 }
@@ -676,7 +662,7 @@ static void list_sort_quicksort(list_t *l, int versus,
     if (pivotid > 0) {
         /* prepare wrapped args, then start thread */
         if (l->threadcount < SIMCLIST_MAXTHREADS-1) {
-            struct list_sort_wrappedparams *wp = malloc(sizeof(struct list_sort_wrappedparams));
+            struct list_sort_wrappedparams *wp = (struct list_sort_wrappedparams *)malloc(sizeof(struct list_sort_wrappedparams));
             l->threadcount++;
             traised = 1;
             wp->l = l;
@@ -709,7 +695,7 @@ int list_iterator_start(list_t *l) {
     if (l->iter_active) return 0;
     l->iter_pos = 0;
     l->iter_active = 1;
-    l->iter_curentry = l->head;
+    l->iter_curentry = l->head_sentinel->next;
     return 1;
 }
 
@@ -744,13 +730,13 @@ list_hash_t list_hash(const list_t *l) {
     hash = l->numels * 2 + 100;
     if (l->attrs.hasher == NULL) {
         /* only use element references */
-        for (x = l->head; x != NULL; x = x->next) {
+        for (x = l->head_sentinel->next; x != l->tail_sentinel; x = x->next) {
             hash += (hash ^ (list_hash_t)x->data);
             hash += hash % l->numels;
         }
     } else {
         /* hash each element with the user-given function */
-        for (x = l->head; x != NULL; x = x->next) {
+        for (x = l->head_sentinel->next; x != l->tail_sentinel; x = x->next) {
             hash += hash ^ l->attrs.hasher(x->data);
             hash += hash % l->numels;
         }
@@ -768,14 +754,8 @@ int list_drop_elem(list_t *l, struct list_entry_s *tmp, unsigned int pos) {
         if (pos < l->numels/2) l->mid = l->mid->next;
     }
     
-    if (tmp->prev != NULL)
-        tmp->prev->next = tmp->next;
-    else    /* be head */
-        l->head = tmp->next;
-    if (tmp->next != NULL)
-        tmp->next->prev = tmp->prev;
-    else    /* be tail */
-        l->tail = tmp->prev;
+    tmp->prev->next = tmp->next;
+    tmp->next->prev = tmp->prev;
 
     /* free what's to be freed */
     if (l->attrs.copy_data && tmp->data != NULL)
@@ -807,7 +787,7 @@ SIMCLIST_NUMBER_COMPARATOR(uint64_t);
 SIMCLIST_NUMBER_COMPARATOR(float);
 SIMCLIST_NUMBER_COMPARATOR(double);
 
-int list_comparator_string(const void *a, const void *b) { return strcmp(b, a); }
+int list_comparator_string(const void *a, const void *b) { return strcmp((const char *)b, (const char *)a); }
 
 /* ready-made metric functions */
 #define SIMCLIST_METER(type)        size_t list_meter_##type(const void *el) { return sizeof(type); }
@@ -825,7 +805,7 @@ SIMCLIST_METER(uint64_t);
 SIMCLIST_METER(float);
 SIMCLIST_METER(double);
 
-size_t list_meter_string(const void *el) { return strlen(el) + 1; }
+size_t list_meter_string(const void *el) { return strlen((const char *)el) + 1; }
 
 /* ready-made hashing functions */
 #define SIMCLIST_HASHCOMPUTER(type)    list_hash_t list_hashcomputer_##type(const void *el) { return (list_hash_t)(*(type *)el); }
@@ -861,33 +841,32 @@ list_hash_t list_hashcomputer_string(const void *el) {
 
 #ifndef NDEBUG
 static int list_repOk(const list_t *l) {
-    int ok;
-    unsigned int i;
+    int ok, i;
     struct list_entry_s *s;
 
     ok = (l != NULL) && (
-            /* no elems, head & tail --> NULL */
-            (l->numels != 0 || (l->head == NULL && l->mid == NULL && l->tail == NULL))
-            /* 1 elem => head == tail both !NULL */
-            && (l->numels != 1 || (l->head != NULL && l->mid == l->head && l->tail == l->head))
+            /* head/tail checks */
+            (l->head_sentinel != NULL && l->tail_sentinel != NULL) &&
+                (l->head_sentinel != l->tail_sentinel) && (l->head_sentinel->prev == NULL && l->tail_sentinel->next == NULL) &&
+            /* empty list */
+            (l->numels > 0 || (l->mid == NULL && l->head_sentinel->next == l->tail_sentinel && l->tail_sentinel->prev == l->head_sentinel)) &&
+            /* spare elements checks */
+            l->spareelsnum <= SIMCLIST_MAX_SPARE_ELEMS
          );
     
     if (!ok) return 0;
 
     if (l->numels >= 1) {
-        /* head and tail terminating endpoints */
-        ok = (l->head->prev == NULL && l->tail->next == NULL);
-        if (!ok) return 0;
         /* correct referencing */
-        for (i = 0, s = l->head; i < (l->numels-1)/2 && s->next != NULL; i++, s = s->next) {
+        for (i = -1, s = l->head_sentinel; i < (int)(l->numels-1)/2 && s->next != NULL; i++, s = s->next) {
             if (s->next->prev != s) break;
         }
-        ok = (i == (l->numels-1)/2 && l->mid == s);
+        ok = (i == (int)(l->numels-1)/2 && l->mid == s);
         if (!ok) return 0;
         for (; s->next != NULL; i++, s = s->next) {
             if (s->next->prev != s) break;
         }
-        ok = (i == l->numels-1);
+        ok = (i == (int)l->numels && s == l->tail_sentinel);
     }
 
     return ok;
