@@ -2,6 +2,10 @@
 #ifndef SIMCLIST_H
 #define SIMCLIST_H
 
+#ifdef __cplusplus
+extern "C" {
+#endif
+
 #include <sys/types.h>
 
 /**
@@ -66,8 +70,8 @@ struct list_attributes_s {
 
 /** list object */
 typedef struct {
-    struct list_entry_s *head;
-    struct list_entry_s *tail;
+    struct list_entry_s *head_sentinel;
+    struct list_entry_s *tail_sentinel;
     struct list_entry_s *mid;
 
     unsigned int numels;
@@ -598,5 +602,9 @@ list_hash_t list_hashcomputer_double(const void *el);
  */
 list_hash_t list_hashcomputer_string(const void *el);
 
+#ifdef __cplusplus
+}
+#endif
+
 #endif
 
