@@ -4,6 +4,9 @@
 
 #include <syslog.h>
 
+#define SLOG_SYSLOG     5
+#define SLOG_STDERRR    10
+
 int sshguard_log_init();
 int sshguard_log(int prio, char *fmt, ...);
 int sshguard_log_fin();
